#!/usr/bin/env python
# coding: utf8

"""Visual keyboard highly customizable for pygame."""

from .vkeyboard import VKeyboardLayout, VKeyboard
from .vrenderers import VKeyboardRenderer

__version__ = '2.0.9'
